from .public import get_symbols, get_symbols_as_updated, PublicData
from .private import Trader

__author__ = """Djordje Pepic"""
__email__ = 'djordje.m.pepic@gmail.com'
__version__ = '0.1.5'

__all__ = [
    'get_symbols', 'get_symbols_as_updated',
    'PublicData', 'Trader'
]



import os, sys
sys.path.append(os.path.abspath('../btfx_trader'))
from collections import defaultdict
from threading import Thread, Event
from time import time, sleep
from queue import Queue
import pytest
import random
import btfxwss.queue_processor
import requests
from btfx_trader import PublicData, Trader


class TestWSS:
    def __init__(self, key=None, secret=None, log_level=None, tr=None):
        self.trader = tr
        self.key = key
        self.secret = secret
        self.log_level = log_level
        ev = Event()
        Thread(target=lambda: (sleep(1e-3), ev.set())).start()
        self.conn = type('', (), {'connected': ev})
        self.queue_processor = btfxwss.queue_processor.QueueProcessor(
            data_q=Queue(), log_level='CRITICAL'
        )

    def start(self):
        pass

    def stop(self):
        pass

    def authenticate(self):
        pass

    def subscribe_to_ticker(self, ticker):
        pass

    def unsubscribe_from_ticker(self, ticker):
        pass

    def subscribe_to_trades(self, trades):
        pass

    def unsubscribe_from_trades(self, trades):
        pass

    def subscribe_to_candles(self, candles, timeframe=None):
        pass

    def unsubscribe_from_candles(self, candles, timeframe=None):
        pass

    def new_order(self, **kwargs):
        self.trader._update('on',
              [kwargs.pop('id', None) or random.randint(1000000, 9999999),
               '', '', kwargs.get('symbol'),
               '', time() * 1000,
               float(kwargs.get('amount')), float(kwargs.get('amount')), '', '',
               '', '', '',
               'ACTIVE', '', '', float(kwargs.get('price')), 0., '']
        )

    def cancel_order(self, **kwargs):
        _id = kwargs.pop('id')
        order = self.trader.orders[_id]
        self.trader._update('oc', [
            _id, '', '', 't' + order['symbol'], '',
            order['timestamp'] * 1000, order['remaining'], order['amount'],
            '', '', '', '', '',  'CANCELLED', '', '',
            order['price'], order['executed_price'], ''
        ])


@pytest.fixture
def patched_get_symbols(monkeypatch):
    got_once = False
    first = b'["btcusd","ethusd","ltcusd","ethltc"]'
    second = b'["btcusd","ethusd","ltcusd","ethltc","sanusd"]'

    def fake_get(*args, **kwargs):
        nonlocal got_once
        if not got_once:
            got_once = True
            d = first
        else:
            d = second
        r = requests.Response()
        r.status_code, r._content = 200, d
        return r

    monkeypatch.setattr('requests.get', fake_get)
    yield


@pytest.fixture
def data():
    t = PublicData(
        types=['tickers', 'trades', ('candles', '1m')],
        symbols=['BTCUSD', 'LTCUSD', 'ETHUSD']
    )
    t.wss = TestWSS('test_key', 'test_secret', log_level='CRITICAL')
    return t


@pytest.fixture
def trader():
    t = Trader('test_key', 'test_secret')
    t.wss = TestWSS('test_key', 'test_secret', log_level='CRITICAL', tr=t)
    return t


@pytest.fixture
def setup_trader():
    def wrapped(t):
        _wallets = [['EXCHANGE', s.lower().replace('usd', ''),
                     0., None, None] for s in t.symbols]
        t._update('ws', [['EXCHANGE', 'usd', 10000., None, None]] + _wallets)
        t._prices = defaultdict(float)
        t._prices.update({
            'BTCUSD': 15000,
            'ETHUSD': 700,
            'LTCUSD': 300,
            'DSHUSD': 1000,
            'XRPUSD': 0.2
        })
    return wrapped


@pytest.fixture
def new_order():
    def wrapper(tr, symbol, price, amount, ts=None, _id=None):
        _id = _id or random.randint(1000000, 9999999)
        tr._update('on',
             [_id, '', '', 't'+symbol, '', (ts or time())*1000, amount, amount,
              '', '', '', '', '',  'ACTIVE', '', '', price, 0., '']
        )
        return _id
    return wrapper


@pytest.fixture
def update_order():
    # This is bad but the alternative is a mock Bitfinex trading platform so...
    def wrapper(tr, _id, price, amount, cancel=False, execute=False, ts=None):
        if execute or cancel:
            cmd = 'oc'
            if execute:
                status = 'EXECUTED @ ($%.2f)' % price
            else:
                status = 'CANCELLED'
        else:
            cmd = 'ou'
            if amount != 0:
                status = 'PARTIALLY EXECUTED'
            else:
                status = 'ACTIVE'

        exc_price = tr._orders[_id]['executed_price']
        if amount != 0:
            if exc_price != 0:
                exc_price = (exc_price + price) / 2

            s = tr._orders[_id]['symbol'].lower().replace('usd', '')
            dollars = tr._wallets['usd'] - amount * price
            tr._update('wu', ['EXCHANGE', 'usd', dollars, None, None])
            tr._update('wu', [
                'EXCHANGE', s, tr._wallets[s] + tr._orders[_id]['executed'] + amount, None, None
            ])

        tr._update(cmd, [_id, '', '', 't' + tr._orders[_id]['symbol'], '',
                         (ts or time()) * 1000,
                         tr._orders[_id]['remaining'] - amount,
                         tr._orders[_id]['amount'],
                         '', '', '', '', '',  status, '', '',
                         price, exc_price, ''])
    return wrapper



from time import time, sleep
from threading import Thread
import pytest


# ========================== #
#       Account tests        #
# ========================== #

def test_connect_close(trader):
    trader.connect()
    sleep(1e-4)
    assert trader.wss.conn.connected.is_set(), \
        'Trader connection not set after connect'
    trader.subscribe('BTCUSD')
    assert 'BTCUSD' in trader.symbols, 'Did not add symbol on subscribe'
    trader.close()
    sleep(1e-4)
    assert trader._disconnect_event.is_set(), \
        'Trader did not set disconnect event on close'


def test_wallets_update(trader):
    # wallet init
    trader._update('ws', [
        ['EXCHANGE', 'usd', 100., None, None],
        ['EXCHANGE', 'btc', 10., None, None],
        ['EXCHANGE', 'eth', 5., None, None]
    ])
    sleep(0.002)  # wait for thread to update
    assert trader.wallets['usd'] == 100., 'Did not update usd in wallet'
    assert trader.wallets['btc'] == 10., 'Did not update btc in wallet'
    assert trader.wallets['eth'] == 5., 'Did not update eth in wallet'
    # wallet updates
    trader._update('wu', ['EXCHANGE', 'usd', 50., None, None])
    trader._update('wu', ['EXCHANGE', 'btc', 20., None, None])
    sleep(0.004)  # wait for thread to update
    assert trader.wallets['usd'] == 50., 'Did not update usd in wallet'
    assert trader.wallets['btc'] == 20., 'Did not update btc in wallet'
    assert trader.wallets['eth'] == 5., 'Did not update eth in wallet'


def test_orders_init_execute(trader):
    # orders init
    ts = time() * 1000
    expected_json_init = dict(
            symbol='BTCUSD', price=50., executed_price=45., amount=10.,
            executed=5., remaining=5., status='PARTIALLY', timestamp=ts / 1000,
        )
    trader._update('os', [
        [123, '', '', 'tBTCUSD', '', ts, 5., 10., '', '', '', '', '',  'PARTIALLY EXECUTED', '', '', 50., 45., '']
    ])
    sleep(0.002)  # wait for thread to update
    assert trader.orders[123] == expected_json_init, 'Incorrect order on init'
    assert trader.orders[123] == list(trader.orders.values())[-1], 'Incorrect latest order'
    assert len(trader.orders) == 1, 'Incorrect length of orders'
    # execute order
    ts = time() * 1000
    expected_json_execute = dict(
            symbol='BTCUSD', price=50., executed_price=43., amount=10.,
            executed=10., remaining=0., status='EXECUTED', timestamp=ts / 1000,
        )
    trader._update(
        'oc',
        [123, '', '', 'tBTCUSD', '', ts, 0., 10., '', '', '', '', '',  'EXECUTED @', '', '', 50., 43., '']
    )
    sleep(0.002)  # wait for thread to update
    assert 123 not in trader.orders, 'Did not remove executed order'
    assert len(trader.orders) == 0, 'Incorrect length of orders'
    assert list(trader.executed_orders.items())[-1] == (123, expected_json_execute), \
        'Incorrect order json added to executed orders'


def test_orders_update(trader, new_order, update_order):
    # new order
    ts = round(time(), 2)
    expected_json_new = dict(
            symbol='BTCUSD', price=50., executed_price=0., amount=10.,
            executed=0., remaining=10., status='ACTIVE', timestamp=ts,
        )
    _id = new_order(trader, 'BTCUSD', 50., 10., ts=ts)
    sleep(0.002)  # wait for thread to update
    assert trader.orders[_id] == expected_json_new, 'Incorrect new order'
    assert trader.orders[_id] == list(trader.orders.values())[-1], 'Incorrect latest order'
    assert len(trader.orders) == 1, 'Incorrect length of orders'
    # update order
    ts = round(time(), 2)
    expected_json_update = dict(
            symbol='BTCUSD', price=25., executed_price=0., amount=10.,
            executed=0., remaining=10., status='ACTIVE', timestamp=ts,
        )
    update_order(trader, _id, 25, 0., ts=ts)
    sleep(0.002)  # wait for thread to update
    assert trader.orders[_id] == expected_json_update, 'Incorrect update order'
    assert trader.orders[_id] == list(trader.orders.values())[-1], 'Incorrect latest order'
    assert len(trader.orders) == 1, 'Incorrect length of orders'
    # cancel order
    update_order(trader, _id, 25, 0, cancel=True)
    sleep(0.002)  # wait for thread to update
    assert _id not in trader.orders, 'Did not remove cancelled order'
    assert len(trader.orders) == 0, 'Incorrect length of orders'


def test_available_balances(trader, setup_trader, new_order, update_order):
    setup_trader(trader)
    trader._wallets['btc'] = 0.1
    trader._wallets['eth'] = 1.
    _id = new_order(trader, 'BTCUSD', 15000, 0.01)

    assert trader.available_balances['usd'] == 9850., \
        'Incorrect wallet value after buy BTCUSD'
    assert trader.available_balances['btc'] == 0.1, \
        'Incorrect wallet value after buy BTCUSD'

    update_order(trader, _id, 15000, 0.01, execute=True)

    assert trader.available_balances['usd'] == 9850., \
        'Incorrect wallet value after execute buy BTCUSD'
    assert trader.available_balances['btc'] == 0.11, \
        'Incorrect wallet value after execute buy BTCUSD'

    _id = new_order(trader, 'ETHUSD', 700, -1.)

    assert trader.available_balances['usd'] == 9850., \
        'Incorrect wallet value after sell ETHUSD'
    assert trader.available_balances['eth'] == 0., \
        'Incorrect wallet value after sell ETHUSD'

    update_order(trader, _id, 700, -1., execute=True)

    assert trader.available_balances['usd'] == 10550., \
        'Incorrect wallet value after execute sell BTCUSD'
    assert trader.available_balances['eth'] == 0., \
        'Incorrect wallet value after execute sell ETHUSD'


def test_position(trader, setup_trader):
    setup_trader(trader)
    assert trader.positions['BTCUSD'] == 0., 'Incorrect position with no coin'
    trader._wallets['usd'] -= 5000.
    trader._wallets['btc'] += 1 / 3
    assert trader.positions['BTCUSD'] == 0.5, 'Incorrect position with balanced dollars and coin'
    trader._wallets['usd'] -= 5000.
    trader._wallets['btc'] += 1 / 3
    assert trader.positions['BTCUSD'] == 1., 'Incorrect position with no dollars'
    trader._wallets['usd'] += 10000.
    trader._wallets['btc'] -= 2 / 3
    assert trader.positions['BTCUSD'] == 0., 'Incorrect position with imbalanced dollars and coin'


def test_value(trader, setup_trader, new_order, update_order):
    setup_trader(trader)
    assert trader.value == 10000., 'Incorrect initial value'
    _id = new_order(trader, 'BTCUSD', 15000, 0.01)
    assert trader.value == 10000., 'Incorrect value during trade'
    update_order(trader, _id, 15000, 0.01, execute=True)
    assert trader.value == 10000., 'Incorrect value after neutral trade'
    trader._prices['BTCUSD'] = 30000
    assert trader.value == 10150., 'Incorrect value after price increase'


def test_value_after_price_update(patched_get_symbols, trader, setup_trader):
    setup_trader(trader)
    trader._wallets['btc'] = 0.1
    trader.connect()
    _id = trader.order('BTCUSD', 15000, dollar_amount=150)
    value = trader.value
    trader.wss.queue_processor.account['Order Cancel'].put(
        [['oc', [_id, '', '', 'tBTCUSD', '', time() * 1000, 0, 0.01,
                 '', '', '', '', '', 'CANCELLED', '', '', 15000, 0, '']
          ], 0]
    )
    trader.wss.queue_processor.tickers[('ticker', 'BTCUSD')].put(
        [[[
            7500.0, 1., 7501.0, 1., -1.0, -0.01,
            7429.8, 26839.8, 7654.1, 7318.2
        ]], 1527273347.1]
    )
    now, diff = time(), 0
    while trader.value == value and diff < 1e-3:
        sleep(1e-5)
        diff = time() - now
    trader.close()
    assert trader.value != value, 'Did not update value in allotted time'


# ========================== #
#        Order tests         #
# ========================== #

def test_cancel_single(trader, setup_trader, new_order):
    setup_trader(trader)
    _id = new_order(trader, 'BTCUSD', 15000, 0.01)
    new_order(trader, 'BTCUSD', 15000, 0.01)
    assert len(trader.orders) == 2, 'Did not add both orders'
    trader.cancel(_id)
    assert len(trader.orders) == 1, 'Did not cancel order'


def test_cancel_all(trader, setup_trader, new_order):
    setup_trader(trader)
    now = time()
    for i in range(10):
        new_order(trader, 'BTCUSD', 15000, 0.01, ts=now - i * 10)
    assert len(trader.orders) == 10, 'Did not add all orders'
    trader.cancel_all(older_than=50)
    assert len(trader.orders) == 5, 'Did not cancel all orders'


def test_wait_execution(trader, setup_trader, new_order, update_order):
    setup_trader(trader)
    _id = new_order(trader, 'BTCUSD', 15000, 0.01)
    with pytest.raises(TimeoutError):
       trader.wait_execution(_id, seconds=0.001)
    del trader._orders[_id]
    _id = new_order(trader, 'BTCUSD', 15000, 0.01)
    result = []

    def wait(_idd, _wait):
        result.append(trader.wait_execution(_idd, seconds=_wait))
    t = Thread(target=wait, args=(_id, 0.1))
    t.start()
    update_order(trader, _id, 15150, 0.01, execute=True)
    t.join()
    assert len(trader.executed_orders) == 1, 'Did not execute order'
    assert list(trader.executed_orders.items())[-1] == (_id, result[0]), 'Incorrect items in orders'


def test_order_multiple_args_fails(trader, setup_trader):
    setup_trader(trader)
    with pytest.raises(AssertionError):
        trader.order("BTCUSD", 15000., dollar_amount=1000, ratio=0.5)


def test_order_dollar_amount(trader, setup_trader, update_order):
    setup_trader(trader)
    symbol, price = "BTCUSD", 15000.
    # buy
    _id = trader.order(symbol, price, dollar_amount=1000)
    assert round(trader.available_balances['usd'], 2) == 9000, \
        'Did not buy correct amount'
    amount = trader.orders[_id]['amount']
    update_order(trader, _id, price, amount, execute=True)
    assert round(trader.available_balances['btc'], 8) == amount, \
        'Did not buy correct amount'
    # sell
    _id = trader.order(symbol, price, dollar_amount=-1000)
    assert round(trader.available_balances['btc'], 8) == 0, \
        'Did not sell correct amount'
    amount = trader.orders[_id]['amount']
    update_order(trader, _id, price, amount, execute=True)
    assert round(trader.available_balances['usd'], 2) == 10000, \
        'Did not sell correct amount'


def test_order_ratio(trader, setup_trader, update_order):
    setup_trader(trader)
    symbol, price = "BTCUSD", 15000.
    # buy
    _id = trader.order(symbol, price, ratio=0.5)
    assert round(trader.available_balances['usd'], 2) == 5000, \
        'Did not buy correct amount'
    amount = trader.orders[_id]['amount']
    update_order(trader, _id, price, amount, execute=True)
    assert round(trader.available_balances['btc'], 8) == amount, \
        'Did not buy correct amount'
    # sell
    _id = trader.order(symbol, price, ratio=-0.5)
    assert round(trader.available_balances['btc'], 7) == round(amount / 2, 7), \
        'Did not sell correct amount'
    amount = trader.orders[_id]['amount']
    update_order(trader, _id, price, amount, execute=True)
    assert round(trader.available_balances['usd'], 2) == 7500, \
        'Did not sell correct amount'
    _id = trader.order(symbol, price, ratio=-1)
    assert round(trader.available_balances['btc'], 8) == 0, \
        'Did not sell correct amount'
    amount = trader.orders[_id]['amount']
    update_order(trader, _id, price, amount, execute=True)
    assert round(trader.available_balances['usd'], 2) == 10000, \
        'Did not sell correct amount'


def test_order_value_ratio(trader, setup_trader, update_order):
    setup_trader(trader)
    symbol, price = "BTCUSD", 15000.
    # buy
    _id = trader.order(symbol, price, value_ratio=0.5)
    assert round(trader.available_balances['usd'], 2) == 5000, \
        'Did not buy correct amount'
    amount = trader.orders[_id]['amount']
    update_order(trader, _id, price, amount, execute=True)
    assert round(trader.available_balances['btc'], 8) == amount, \
        'Did not buy correct amount'
    # sell
    _id = trader.order(symbol, price, value_ratio=-0.5)
    assert round(trader.available_balances['btc'], 8) == 0, \
        'Did not sell correct amount'
    amount = trader.orders[_id]['amount']
    update_order(trader, _id, price, amount, execute=True)
    assert round(trader.available_balances['usd'], 2) == 10000, \
        'Did not sell correct amount'


def test_order_market(trader, setup_trader, update_order):
    setup_trader(trader)
    # buy
    _id = trader.order("BTCUSD", 'market', dollar_amount=1000)
    assert round(trader.available_balances['usd'], 2) == 9000, \
        'Did not buy correct amount'
    amount = trader.orders[_id]['amount']
    update_order(trader, _id, 15000., amount, execute=True)
    assert round(trader.available_balances['btc'], 8) == amount, \
        'Did not buy correct amount'
    # sell
    _id = trader.order("BTCUSD", 'market', dollar_amount=-1000)
    assert round(trader.available_balances['btc'], 8) == 0, \
        'Did not sell correct amount'
    amount = trader.orders[_id]['amount']
    update_order(trader, _id, 15000., amount, execute=True)
    assert round(trader.available_balances['usd'], 2) == 10000, \
        'Did not sell correct amount'


def test_order_pad_price(trader, setup_trader, update_order):
    setup_trader(trader)
    symbol, price = "DSHUSD", 1000.
    # buy
    _id = trader.order(symbol, price, dollar_amount=1100, pad_price=0.1)
    assert round(trader.available_balances['usd'], 2) == 8900, \
        'Did not buy correct amount'
    amount = trader.orders[_id]['amount']
    update_order(trader, _id, trader.orders[_id]['price'], amount, execute=True)
    assert round(trader.available_balances['dsh'], 8) == amount, \
        'Did not buy correct amount'
    # sell
    _id = trader.order(symbol, price, dollar_amount=-900, pad_price=0.1)
    assert round(trader.available_balances['dsh'], 8) == 0, \
        'Did not sell correct amount'
    amount = trader.orders[_id]['amount']
    update_order(trader, _id, trader.orders[_id]['price'], amount, execute=True)
    assert round(trader.available_balances['usd'], 2) == 9800, \
        'Did not sell correct amount'


def test_order_id_with_previous_executed_orders(trader, setup_trader, update_order):
    setup_trader(trader)
    old_order = trader.wss.new_order
    trader._sleep_time = 1e-6

    def execute_on_order(**kwargs):
        def wrapped():
            sleep(1e-4)
            old_order(cid=int(time()), type='EXCHANGE LIMIT',
                      symbol='tBTCUSD', price=kwargs.get('price'),
                      amount=kwargs.get('amount'), id=123)
            update_order(trader, 123, 15000, 0.01, execute=True)
        Thread(target=wrapped).start()

    trader.wss.new_order = execute_on_order
    trader.order('BTCUSD', 15000, dollar_amount=300)



from queue import Empty
from threading import Timer
from time import sleep
import pytest
from btfx_trader import get_symbols_as_updated

NOWAIT_DATA = [
    (
        'tickers', ('ticker', 'BTCUSD'),
        [([[7500.0, 1., 7501.0, 1., -1.0, -0.01, 7429.8, 26839.8, 7654.1, 7318.2]], 1527273347.1)],
        ['bid'], [7500.0]
    ),
    (
        'trades', ('trades', 'BTCUSD'),
        [
            (['te', [250014166, 1527273332589, -0.5, 7500.0]], 1527273332.8),
            (['tu', [250014166, 1527273332589, -0.5, 7500.0]], 1527273332.8)
        ],
        ['price'], [7500.0]
    ),
    (
        'candles', ('candles', 'BTCUSD', '1m'),
        [([[1527273240000, 7500.0, 7500.8, 7500.0, 7500.5, 0.5]], 1527273240.0)],
        ['open'], [7500.0]
    )
]


def test_get_symbols_as_updated(patched_get_symbols):
    gen = iter(get_symbols_as_updated(check_every=0))
    for expected in ['BTCUSD', 'ETHUSD', 'LTCUSD', 'SANUSD']:
        symbol = next(gen)
        assert symbol == expected, 'Did not get correct symbol from gen'


def test_connect_close(data):
    data.connect()
    sleep(1e-4)
    assert data.wss.conn.connected.is_set(), \
        'Data connection not set after connect'
    data.close()


@pytest.mark.parametrize('name,pair,data_to_put,keys,values', NOWAIT_DATA)
def test_get_nowait(data, name, pair, data_to_put, keys, values):
    for d in data_to_put:
        getattr(data.wss.queue_processor, name)[pair].put_nowait(d)
    if name == 'candles':
        name = ('candles', '1m')
    try:
        d = data.get_nowait(name, 'BTCUSD')
        for k, v in zip(keys, values):
            assert d[k] == v, 'Did not get correct data for %s' % str(name)
    except Empty:
        assert name == 'trades', 'Error getting data'


@pytest.mark.parametrize('name,pair,data_to_put,keys,values', NOWAIT_DATA)
def test_get(data, name, pair, data_to_put, keys, values):
    def put(n, p, dd):
        for _d in dd:
            getattr(data.wss.queue_processor, n)[p].put(_d)

    Timer(1e-4, put, args=[name, pair, data_to_put]).start()
    if name == 'candles':
        name = ('candles', '1m')
    d = data.get(name, 'BTCUSD')
    for k, v in zip(keys, values):
        assert d[k] == v, 'Did not get correct data for %s' % str(name)


def test_iter(data):
    for name, pair, d, _, _ in NOWAIT_DATA:
        for _d in d:
            getattr(data.wss.queue_processor, name)[pair].put_nowait(_d)
    count = 0
    types = {'tickers', 'trades', ('candles', '1m')}
    for _type, symbol, d in data:
        assert symbol == 'BTCUSD', "Incorrect symbol for data iteration"
        assert _type in types, "Incorrect data type for iteration"
        count += 1
        if count >= 2:
            break



from contextlib import suppress
from queue import Queue, Empty
import logging
import time

import requests
import btfxwss
import btfxwss.queue_processor
import btfxwss.connection

log = logging.getLogger(__name__)
# Multiprocessing queue is slow...
btfxwss.connection.Queue = btfxwss.queue_processor.Queue = Queue


def get_symbols():
    return list(
        map(lambda q: q.upper(),
            filter(lambda i: 'usd' in i,
                   requests.get('https://api.bitfinex.com/v1/symbols').json()))
    )


def get_symbols_as_updated(check_every=43200):
    symbols = set(get_symbols())
    for s in list(sorted(symbols)):
        yield s
    while True:
        time.sleep(check_every)
        current_symbols = set(get_symbols())
        new_symbols = current_symbols.difference(symbols)
        symbols.update(new_symbols)
        for s in list(sorted(new_symbols)):
            yield s


def _handle_ticker(data):
    return {
        'bid': data[0][0], 'bid_size': data[0][1],
        'ask': data[0][2], 'ask_size': data[0][3],
        'mid': round((data[0][0] + data[0][2]) / 2, 4),
        'weighted_mid': round((data[0][0] * data[0][1] +
                               data[0][2] * data[0][3]) /
                              (data[0][1] + data[0][3]), 4),
        'last_price': data[0][6],
        'spread': round(2 * (data[0][2] - data[0][0]) /
                        (data[0][0] + data[0][2]), 8),
        'change': data[0][5],
        'volume': data[0][7],
        'high': data[0][8],
        'low': data[0][9],
        'timestamp': round(time.time(), 0)
    }


def _handle_trade(data):
    if data[0] == 'tu':
        return {
            'id': data[1][0],
            'timestamp': data[1][1] / 1000,
            'price': data[1][3],
            'amount': data[1][2]
        }


def _handle_candle(data):
    return {
        'timestamp': data[0][0] / 1000,
        'open': data[0][1],
        'high': data[0][2],
        'low': data[0][3],
        'close': data[0][4],
        'volume': data[0][5]
    }


class PublicData:
    """
    Wrapper for BtfxWss for access to public bitfinex data
    :param types: list:
        (optional) data types to filter data by (tickers, trades, candles)
    :param symbols: list:
        (optional) symbol to filter data by (e.g. ['BTCUSD', 'ETHUSD'])
    """
    candle_periods = [
        '1m', '5m', '15m', '30m',
        '1h', '3h', '6h', '12h',
        '1D', '7D', '14D', '1M'
    ]

    def __init__(self, types=None, symbols=None):
        self.types = set(types or [])
        self.symbols = set(symbols or [])
        self._assert_types_are_correct(self.types)
        self.wss = btfxwss.BtfxWss(key='', secret='', log_level='CRITICAL')

    def __iter__(self):
        while True:
            types, symbols = self.types.copy(), self.symbols.copy()
            for _type in types:
                for symbol in symbols:
                    try:
                        yield _type, symbol, self.get_nowait(_type, symbol)
                    except Empty:
                        continue

    def get(self, _type, symbol):
        """
        Gets next available data (blocks until new datum is received)
        :param _type: str: type of data to get (e.g. 'tickers', 'candles'...)
        :param symbol: str:
            symbol for which to get data (e.g. 'BTCUSD', 'ETHUSD'...)
        :return: dict: requested data in json format
        :raises: AssertionError:
            if either of the provided type and symbol are unrecognized
        """
        assert _type in self.types, "Unknown type '%s'" % _type
        assert symbol in self.symbols, "Unknown symbol '%s'" % symbol
        while True:
            try:
                return self._get_nowait(_type, symbol)
            except Empty:
                time.sleep(1e-6)

    def get_nowait(self, _type, symbol):
        """
        Gets currently available data with no blocking
        :param _type: str: type of data to get (e.g. 'tickers', 'candles'...)
        :param symbol: str:
            symbol for which to get data (e.g. 'BTCUSD', 'ETHUSD'...)
        :return: dict:
            requested data in json format
        :raises:
            AssertionError:
                if either of the provided type and symbol are unrecognized
            Empty:
                if no new data is available (same as queue.Queue api)
        """
        assert _type in self.types, "Unknown type '%s'" % _type
        assert symbol in self.symbols, "Unknown symbol '%s'" % symbol
        return self._get_nowait(_type, symbol)

    def subscribe(self, _type, symbol):
        """
        Subscribe to a stream of data
        :param _type: str: type of data to get (e.g. 'tickers', 'candles'...)
        :param symbol: str:
            symbol for which to get data (e.g. 'BTCUSD', 'ETHUSD'...)
        """
        self._assert_types_are_correct([_type])
        self.types.add(_type)
        self.symbols.add(symbol)

        if _type == 'tickers':
            self.wss.subscribe_to_ticker(symbol)
        elif _type == 'trades':
            self.wss.subscribe_to_trades(symbol)
        elif isinstance(_type, tuple):
            self.wss.subscribe_to_candles(symbol, timeframe=_type[1])

    def connect(self):
        """Open a connection to a Bitfinex websocket"""
        self.wss.start()
        while not self.wss.conn.connected.is_set():
            time.sleep(1e-4)

        for _type in self.types:
            for symbol in self.symbols:
                self.subscribe(_type, symbol)

    def close(self):
        """Close the connection to the Bitfinex websocket"""
        for symbol in self.symbols:
            with suppress(KeyError):
                self.wss.unsubscribe_from_ticker(symbol)
            with suppress(KeyError):
                self.wss.unsubscribe_from_trades(symbol)
            with suppress(KeyError):
                self.wss.unsubscribe_from_candles(symbol)
        self.wss.stop()

    @staticmethod
    def _assert_types_are_correct(types):
        candle_types = []
        for _type in types:
            if isinstance(_type, str):
                assert _type in {'tickers', 'trades'}, \
                    "String type must be either 'tickers' or 'trades'"
            else:
                candle_types.append(_type)

        for name, period in candle_types:
            assert name == 'candles', \
                'Candles types must be in the format ("candles", <period>)'
            assert period in PublicData.candle_periods, \
                'Candle period must be one of %s' % PublicData.candle_periods

    @staticmethod
    def _clean_data(_type, data):
        result = data[0]
        if _type == 'trades':
            result = data[0] if isinstance(data[0][0], str) \
                else ['tu', data[0][0][0]]
        elif isinstance(_type, tuple):
            result = data[0] if isinstance(data[0][0][0], (int, float)) \
                else [data[0][0][0]]
        return result

    def _get_queue(self, _type, symbol):
        if _type == 'tickers':
            pair = ('ticker', symbol)
        elif _type == 'trades':
            pair = ('trades', symbol)
        else:
            pair = ('candles', symbol, _type[1])
            _type = 'candles'
        return getattr(self.wss.queue_processor, _type)[pair]

    def _get_nowait(self, _type, symbol):
        if _type == 'tickers':
            handler = _handle_ticker
        elif _type == 'trades':
            handler = _handle_trade
        else:
            handler = _handle_candle

        log.debug("Getting %s - %s from queues", str(_type), symbol)
        data = self._get_queue(_type, symbol).get_nowait()
        result = handler(self._clean_data(_type, data))
        if result is None:
            log.debug("%s - %s queue is empty", _type, symbol)
            raise Empty

        return result



import logging
import time
from collections import deque, defaultdict, OrderedDict
from contextlib import suppress
from queue import Empty
from threading import Thread, Event

from btfxwss import BtfxWss
from .public import get_symbols_as_updated

log = logging.getLogger(__name__)


def _jsonify_orders(data):
    return [
        (datum[0],
         dict(
             symbol=datum[3][1:].upper(),
             price=datum[16],
             executed_price=datum[17],
             amount=datum[7],
             executed=datum[7] - datum[6],
             remaining=datum[6],
             status=datum[13].split()[0],
             timestamp=datum[5] / 1000
         )) for datum in data]


class _Account:
    """
    Class for parsing live web socket data
    from bitfinex into user-friendly interface
    """
    def __init__(self):
        self._handlers = {
            'o': self._handle_order,
            't': self._handle_ticker,
            'w': self._handle_wallet
        }
        self._executed_orders = deque(maxlen=100)
        self._orders = OrderedDict()
        self._prices = defaultdict(float)

        self._available_balances = defaultdict(float)
        self._wallets = defaultdict(float)

    @property
    def available_balances(self):
        """
        Amount of currency in each wallet currently available for trading
        :return: dict: available trading balances
        """
        balances = self._wallets.copy()
        for order in self._orders.values():
            if order['remaining'] < 0:
                symbol = order['symbol'].lower().replace('usd', '')
                balances[symbol] += order['remaining']
            else:
                balances['usd'] -= order['remaining'] * order['price']
        return defaultdict(float,
                           **{c: round(balances[c], 8) for c in balances})

    @property
    def executed_orders(self):
        """
        Sorted historical executed orders (earliest first) addressed by id
        :return: OrderedDict: orders executed since initialization
        """
        return OrderedDict(self._executed_orders)

    @property
    def orders(self):
        """
        Sorted currently active orders (earliest first) addressed by order id
        :return: OrderedDict: currently active orders
        """
        return self._orders

    @property
    def positions(self):
        """
        Ratio of total value invested in each symbol (excluding other symbols)
        :return: dict: symbol positions
        """
        return defaultdict(float, **{
            symbol: (self._wallets[symbol.lower().replace('usd', '')]
                     * self._prices[symbol]) / self.value
            for symbol in self._prices
        })

    @property
    def value(self):
        """
        Approximate value of all balances and investments
        :return: float: total value
        """
        total = 0.
        for currency, amount in self._wallets.items():
            if currency != 'usd':
                amount *= self._prices[currency.upper() + 'USD']
            total += amount
        return round(total, 2)

    @property
    def wallets(self):
        """
        Wallet balances for all symbols ('usd' is regular currency)
        :return: dict: wallet balances
        """
        return self._wallets

    @staticmethod
    def _log_order(action, _id, symbol, amount, price):
        log.info('%d (%6s) %-7s %-4s, %.8f for $%.2f at $%.2f',
                 _id, symbol, action, 'BUY' if amount > 0 else 'SELL',
                 abs(amount), abs(amount * price), price)

    def _handle_order_update(self, cmd, orders):
        for _id, order in sorted(orders, key=lambda t: t[1]['timestamp']):
            self._orders[_id] = order
            if cmd == 'on':
                self._log_order('SUBMIT', _id, order['symbol'],
                                order['amount'], order['price'])

    def _handle_order_close(self, orders):
        for _id, order in orders:
            if order['status'].startswith('CANCEL'):
                self._log_order('CANCEL', _id, order['symbol'],
                                order['amount'], order['price'])
            else:
                self._executed_orders.append((_id, order))
                self._log_order('EXECUTE', _id, order['symbol'],
                                order['executed'], order['executed_price'])
            if _id in self._orders:
                del self._orders[_id]

    def _handle_order(self, cmd, data):
        if cmd != 'os':
            data = [data]
        orders = _jsonify_orders(data)
        if cmd in {'os', 'on', 'ou'}:
            self._handle_order_update(cmd, orders)
        else:
            self._handle_order_close(orders)

    def _handle_ticker(self, cmd, data):
        symbol = cmd[1:]
        weighted_mid = (data[0][0] * data[0][1] +
                        data[0][2] * data[0][3]) / (data[0][1] + data[0][3])
        self._prices[symbol] = weighted_mid

    def _handle_wallet(self, cmd, data):
        if cmd == 'ws':
            for datum in data:
                if datum[0].lower() == 'exchange':
                    self._wallets[datum[1].lower()] = float(datum[2])
        elif cmd == 'wu':
            self._wallets[data[1].lower()] = float(data[2])

    def _update(self, cmd, data):
        self._handlers[cmd[0]](cmd, data)


class Trader(_Account):
    """
    Class interface for trading on the Bitfinex Exchange
    :param key: str: Bitfinex api-key
    :param secret:  str: Bitfinex api-secret
    """
    _sleep_time = 0.01
    _min_order_value = 35.
    _max_order_history = 100
    _trade_types = {'market', 'limit'}
    _sources = (
        'Orders', 'Order New', 'Order Update', 'Order Cancel', 'Wallets'
    )

    def __init__(self, key, secret):
        super(Trader, self).__init__()
        self.symbols = []
        self.symbols_gen = get_symbols_as_updated()
        self.wss = BtfxWss(key=key, secret=secret, log_level='CRITICAL')
        self._disconnect_event = Event()
        self._receiver = None

    def cancel(self, _id):
        """
        Cancel an order
        :param _id: int: order id
        :return: None
        """
        self.wss.cancel_order(multi=False, id=_id)

    def cancel_all(self, older_than=0):
        """
        Cancel all orders older than a certain time
        :param older_than: int/float:
            age of order that should be cancelled (in seconds)
        :return: None
        """
        now = time.time()
        for _id, order in self._orders.copy().items():
            if now - order['timestamp'] > older_than:
                self.cancel(_id)

    def connect(self):
        """Open a connection to a Bitfinex websocket"""
        self.wss.start()
        while not self.wss.conn.connected.is_set():
            time.sleep(1e-4)
        self.wss.authenticate()
        # This thread will wait for 99.999% of its life
        subscriber_thread = Thread(target=self._subscribe)
        subscriber_thread.setDaemon(True)
        subscriber_thread.start()
        # This thread should be working and is therefore joined
        self._receiver = Thread(target=self._receive)
        self._receiver.start()

    def close(self):
        """Close the connection to the Bitfinex websocket"""
        self._disconnect_event.set()
        for symbol in self.symbols:
            self.wss.unsubscribe_from_ticker(symbol)
        self.wss.stop()
        self._receiver.join()

    def order(self, symbol, price, *,
              dollar_amount=None, ratio=None, value_ratio=None,
              trade_type='limit', pad_price=None, return_id=True):
        """
        Make an exchange order
        :param symbol: str:
            ticker symbol to order (e.g. 'BTCUSD', 'ETHUSD'...)

        :param price: float:
            Price at which to submit order
            You can also pass "market" as an argument for price
            to order at current market value
            e.g. Trader.order('BTCUSD', 'market', dollar_amount=5)

        :param dollar_amount: float:
            Dollar equivalent value of trade amount (negative for selling)
            e.g. ordering 10 BTCUSD at $5 per coin
                -> Trader.order('BTCUSD', 5, dollar_amount=50)

        :param ratio: float:
            Ratio of available balance for requested symbol (negative for sell)
            e.g. With $1000 USD in wallet, ordering 100 BTCUSD at $5 per coin
                -> Trader.order('BTCUSD', 5, ratio=0.5)

        :param value_ratio: float:
            Ratio of Trader.value (negative for selling)
            e.g. With $500 in USD and $500 in BTCUSD (at $5) in wallet,
                 ordering 100 BTCUSD at $5 per coin
                -> Trader.order('BTCUSD', 5, value_ratio=0.5)

        :param trade_type: str: (optional)
            Bitfinex api trade type - one of {'market', 'limit'}
            see https://www.bitfinex.com/features for details

        :param pad_price: float: (optional)
            Ratio based price padding - used to undercut or overshoot price
            If buying then pad_price * price is added to submitted price
            If selling then pad_price * price is subtracted from price

        :param return_id: bool: (optional)
            If true, Trader.order blocks until order_id is returned
            from Bitfinex, otherwise returns immediately

        :return: int:
            If return_id is True, then order id is returned, otherwise None

        :raises: AssertionError:
            If values passed are non-consistent
        """

        assert dollar_amount or ratio or value_ratio, \
            'Must provide either `dollar_amount`, `ratio` or `value_ratio`'
        assert sum(bool(i) for i in [dollar_amount, ratio, value_ratio]) == 1,\
            'Must provide only 1 of `dollar_amount`, `ratio` or `value_ratio`'
        assert trade_type.lower() in self._trade_types, \
            'Unknown trade type, try one of these: %s' % str(self._trade_types)

        symbol_lower = symbol.lower().replace('usd', '')
        symbol = symbol.upper().split('USD')[0] + 'USD'
        buying = (dollar_amount or ratio or value_ratio) > 0

        if price == 'market' or trade_type == 'market':
            trade_type = 'market'
            price = self._prices[symbol]

        if pad_price:
            delta = price * pad_price
            price += max(delta, 0.01) if buying else min(-delta, -0.01)
        assert price >= 0.01, 'Price cannot be less than $0.01'

        if buying:
            max_amount = self.available_balances['usd'] / price
        else:
            max_amount = self.available_balances[symbol_lower]

        if dollar_amount:
            amount = dollar_amount / price
        elif ratio:
            amount = max_amount * ratio
        else:
            amount = min(max(-max_amount, self.value * value_ratio / price),
                         max_amount)
        amount, max_amount = round(amount, 8), round(max_amount, 8)

        assertion_msg = 'Trade value (${:.2f}) is %s available trade value ' \
                        '($%.2f)'.format(abs(amount * price))
        assert abs(amount) >= self._min_order_value / price, \
            assertion_msg % ('below minimum', self._min_order_value)
        assert abs(amount) <= max_amount, \
            assertion_msg % ('above maximum', max_amount * price)

        current_order_ids = set(self._orders.keys())
        if len(self._executed_orders) == 0:
            last_executed_id = None
        else:
            last_executed_id = self._executed_orders[-1][0]

        self.wss.new_order(
            cid=int(time.time()),
            type="EXCHANGE %s" % trade_type.upper(),
            symbol="t%s" % symbol,
            amount="%.8f" % amount,
            price="%.2f" % price,
        )

        while return_id:
            for _id in self._orders.copy().keys():
                # new order arrives in _orders
                if _id not in current_order_ids:
                    return _id
            else:
                # new order is executed immediately
                if len(self._executed_orders) > 0 \
                 and self._executed_orders[-1][0] != last_executed_id:
                    return self._executed_orders[-1][0]
                time.sleep(self._sleep_time)

    def subscribe(self, symbol):
        """
        Subscribe to a symbol for price watching
        :param symbol: str: symbol to subscribe to (e.g. 'BTCUSD', 'ETHUSD'...)
        """
        self.symbols.append(symbol)
        self.wss.subscribe_to_ticker(symbol)

    def wait_execution(self, _id, seconds=1e9):
        """
        Waits for an order to execute
        :param _id: int: id of order to wait for
        :param seconds: int/float: time to wait before raising error
        :return: dict: order in json format
        :raises: TimeoutError (if order is not executed in given time)
        """
        start_time = time.time()
        while time.time() - start_time < seconds:
            try:
                ids, trades = tuple(zip(*self._executed_orders))
                return trades[ids.index(_id)]
            except ValueError:
                time.sleep(self._sleep_time)
        raise TimeoutError('Waiting for execution of order '
                           '%d timed out after %d seconds' % (_id, seconds))

    def _receive(self):
        while not self._disconnect_event.is_set():
            for source in self._sources:
                with suppress(Empty):
                    q = self.wss.queue_processor.account[source]
                    cmd, data = q.get_nowait()[0]
                    self._update(cmd, data)

            for symbol in self.symbols:
                with suppress(Empty):
                    q = self.wss.queue_processor.tickers[('ticker', symbol)]
                    data = q.get_nowait()[0]
                    self._update('t' + symbol, data)

    def _subscribe(self):
        for symbol in self.symbols_gen:
            self.subscribe(symbol)
